/*
 * USB ICM532 WebCam driver
 * Copyright (c) 2004 Fabio Di Niro
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * This source code is based heavily on the vicam webcam driver which was
 * written by Joe Burks, Christopher L Cheney, Pavel Machek, John Tyner and
 * Monroe Williams
 *
 */

/* internal rev. 0.2 */
 
#include <linux/kernel.h>
#include <linux/wrapper.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/videodev.h>
#include <linux/usb.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include "/usr/src/linux/drivers/usb/usbvideo.h"

#ifndef bool
#define bool int
#endif



/* 
 * Version Information 
 */
#define DRIVER_VERSION "v0.2"
#define DRIVER_AUTHOR "Fabio Di Niro"
#define DRIVER_DESC "ICM532 WebCam Driver"

/* 
 * Debug
 */
#define ICM532_DEBUG

#ifdef ICM532_DEBUG
#define ADBG(lineno,fmt,args...) printk(fmt, jiffies, __FUNCTION__, lineno, ##args)
#define DBG(fmt,args...) ADBG((__LINE__),KERN_DEBUG __FILE__"(%ld):%s (%d):"fmt,##args)
#else
#define DBG(fmn,args...) do {} while(0)
#endif

#define ICM532_BYTES_PER_PIXEL 3
// max received raw frame size = 5 urbs of 32 packets of 1023 bytes each
#define ICM532_MAX_READ_SIZE (1023*32)
#define ICM532_MAX_FRAME_SIZE (ICM532_BYTES_PER_PIXEL*352*288)
#define ICM532_FRAMES 2 // TODO

/* 
 * icm532 init urb type
 */

#define	WRITE_URB		0
#define	READ_URB		1
#define WAIT_VALUE  	2
#define SET_INTERFACE	3
#define RESET_PIPE		4

typedef struct {
    u_int8_t  	action;
    u_int8_t 	icm_register;
    u_int8_t 	value;
} icm532_urb_t;

/*
 * icm532 init urb definition
 * data structure for regs and values come from an idea of 
 * Takafumi Mizuno
 */
#define INIT_URBS 	85
icm532_urb_t icm532_init_sequence[] = {
    { WRITE_URB, 	0x39, 0xb0 }, /* #005 */
			
	{ READ_URB, 	0x0c, 0xf4 }, /* #006 -----------------*/
	{ READ_URB, 	0x0d, 0x01 }, /* #007        |         */
	{ READ_URB, 	0x28, 0x63 }, /* #008        |         */
	{ READ_URB, 	0x29, 0x07 }, /* #009        |         */
	{ READ_URB, 	0x2c, 0x48 }, /* #010        |         */
	{ READ_URB, 	0x2d, 0x01 }, /* #011        |         */
	{ READ_URB, 	0x2e, 0x06 }, /* #012    seq read 1    */
	{ READ_URB, 	0x2f, 0x96 }, /* #013        |         */
	{ READ_URB, 	0x30, 0x04 }, /* #014        |         */
	{ READ_URB, 	0x2a, 0x73 }, /* #015        |         */
	{ READ_URB, 	0x2b, 0x73 }, /* #016        |         */
	{ READ_URB, 	0x35, 0x09 }, /* #017        |         */
	{ READ_URB, 	0x34, 0x23 }, /* #018        |         */
	{ READ_URB, 	0x83, 0x43 }, /* #019 -----------------*/

	{ WRITE_URB, 	0x01, 0x30 }, /* #020 -----------------*/
	{ WRITE_URB, 	0x03, 0x00 }, /* #021        |         */
	{ WRITE_URB, 	0x04, 0x01 }, /* #022        |         */
	{ WRITE_URB, 	0x05, 0x00 }, /* #023        |         */
	{ WRITE_URB, 	0x06, 0x01 }, /* #024        |         */
	{ WRITE_URB, 	0x08, 0x00 }, /* #025        |         */
	{ WRITE_URB, 	0x03, 0x01 }, /* #026        |         */
	{ WRITE_URB, 	0x04, 0x11 }, /* #027        |         */
	{ WRITE_URB, 	0x05, 0x80 }, /* #028        |         */
	{ WRITE_URB, 	0x06, 0x01 }, /* #029        |         */
	{ WRITE_URB, 	0x08, 0x00 }, /* #030        |         */
	{ WRITE_URB, 	0x03, 0x02 }, /* #031        |         */
	{ WRITE_URB, 	0x04, 0x14 }, /* #032        |         */
	{ WRITE_URB, 	0x05, 0x00 }, /* #033        |         */
	{ WRITE_URB, 	0x06, 0x05 }, /* #034    cam |         */
	{ WRITE_URB, 	0x08, 0x00 }, /* #035        |         */
	{ WRITE_URB, 	0x03, 0x03 }, /* #036   initialization */
	{ WRITE_URB, 	0x04, 0x1c }, /* #037        |         */
	{ WRITE_URB, 	0x05, 0x00 }, /* #038        |         */
	{ WRITE_URB, 	0x06, 0x05 }, /* #039        |         */
	{ WRITE_URB, 	0x08, 0x00 }, /* #040        |         */
	{ WRITE_URB, 	0x03, 0x04 }, /* #041        |         */
	{ WRITE_URB, 	0x04, 0x1e }, /* #042        |         */
	{ WRITE_URB, 	0x05, 0x00 }, /* #043        |         */
	{ WRITE_URB, 	0x06, 0x0d }, /* #044        |         */
	{ WRITE_URB, 	0x08, 0x00 }, /* #045        |         */
	{ WRITE_URB, 	0x03, 0x05 }, /* #046        |         */
	{ WRITE_URB, 	0x04, 0x1f }, /* #047        |         */
	{ WRITE_URB, 	0x05, 0x00 }, /* #048        |         */
	{ WRITE_URB, 	0x06, 0x05 }, /* #049        |         */
	{ WRITE_URB, 	0x08, 0x00 }, /* #050        |         */
	{ WRITE_URB, 	0x03, 0x06 }, /* #051        |         */
	{ WRITE_URB, 	0x04, 0x19 }, /* #052        |         */
	{ WRITE_URB, 	0x05, 0x05 }, /* #053        |         */
	{ WRITE_URB, 	0x06, 0x05 }, /* #054        |         */
	{ WRITE_URB, 	0x08, 0x00 }, /* #055        |         */
	{ WRITE_URB, 	0x03, 0x07 }, /* #056        |         */
	{ WRITE_URB, 	0x04, 0x1b }, /* #057        |         */
	{ WRITE_URB, 	0x05, 0x01 }, /* #058        |         */
	{ WRITE_URB, 	0x06, 0x05 }, /* #059        |         */
	{ WRITE_URB, 	0x08, 0x00 }, /* #060        |         */
	{ WRITE_URB, 	0x03, 0x08 }, /* #061        |         */
	{ WRITE_URB, 	0x04, 0x1e }, /* #062        |         */
	{ WRITE_URB, 	0x05, 0x09 }, /* #063        |         */
	{ WRITE_URB, 	0x06, 0x05 }, /* #064        |         */
	{ WRITE_URB, 	0x08, 0x00 }, /* #065        |         */
	{ WRITE_URB, 	0x03, 0x09 }, /* #066        |         */
	{ WRITE_URB, 	0x04, 0x2e }, /* #067        |         */
	{ WRITE_URB, 	0x05, 0x89 }, /* #068        |         */
	{ WRITE_URB, 	0x06, 0x0d }, /* #069        |         */
	{ WRITE_URB, 	0x08, 0x00 }, /* #070        |         */
	{ WRITE_URB, 	0x03, 0x0a }, /* #071        |         */
	{ WRITE_URB, 	0x04, 0x2f }, /* #072        |         */
	{ WRITE_URB, 	0x05, 0x89 }, /* #073        |         */
	{ WRITE_URB, 	0x06, 0x05 }, /* #074        |         */
	{ WRITE_URB, 	0x08, 0x00 }, /* #075        |         */
	{ WRITE_URB, 	0x03, 0x0b }, /* #076        |         */
	{ WRITE_URB, 	0x04, 0xd9 }, /* #077        |         */
	{ WRITE_URB, 	0x05, 0x0d }, /* #078        |         */
	{ WRITE_URB, 	0x06, 0x05 }, /* #079        |         */
	{ WRITE_URB, 	0x08, 0x00 }, /* #080        |         */
	{ WRITE_URB, 	0x03, 0x0c }, /* #081        |         */
	{ WRITE_URB, 	0x04, 0xf1 }, /* #082        |         */
	{ WRITE_URB, 	0x05, 0x09 }, /* #083        |         */
	{ WRITE_URB, 	0x06, 0x05 }, /* #084        |         */
	{ WRITE_URB, 	0x08, 0x00 }, /* #085        |         */
	{ WRITE_URB, 	0x07, 0x0d }, /* #086        |         */
	{ WRITE_URB, 	0x01, 0x29 }, /* #087        |         */
	{ WRITE_URB, 	0x91, 0x32 }, /* #088        |         */
	{ WRITE_URB, 	0x94, 0x00 }  /* #089 -----------------*/
};

#define START_CAPTURE_URBS 	140 // was 141
icm532_urb_t icm532_start_capture_sequence[] = {
    { SET_INTERFACE,0x00, 0x07 }, /* #090                   */
	{ READ_URB, 	0x0c, 0xf4 }, /* #091 -----------------*/
	{ READ_URB, 	0x0d, 0x01 }, /* #092        |         */
	{ READ_URB, 	0x28, 0x63 }, /* #093        |         */
	{ READ_URB, 	0x29, 0x07 }, /* #094        |         */
	{ READ_URB, 	0x2c, 0x48 }, /* #095        |         */
	{ READ_URB, 	0x2d, 0x01 }, /* #096        |         */
	{ READ_URB, 	0x2e, 0x06 }, /* #097    seq read 1    */
	{ READ_URB, 	0x2f, 0x96 }, /* #098        |         */
	{ READ_URB, 	0x30, 0x04 }, /* #099        |         */
	{ READ_URB, 	0x2a, 0x73 }, /* #100        |         */
	{ READ_URB, 	0x2b, 0x73 }, /* #101        |         */
	{ READ_URB, 	0x35, 0x09 }, /* #102        |         */
	{ READ_URB, 	0x34, 0x23 }, /* #103        |         */
	{ READ_URB, 	0x83, 0x43 }, /* #104 -----------------*/
			
	{ WRITE_URB, 	0x0e, 0x49 }, /* #105 AD_HEIGHTL       */
	{ WRITE_URB, 	0x0f, 0x03 }, /* #105 AD_HEIGHTH       */
	
	{ WRITE_URB, 	0x1c, 0x20 }, /* #106 AD_EXPOSE_TIMEL  */
	{ WRITE_URB, 	0x1d, 0x03 }, /* #106 AD_EXPOSE_TIMEH  */
	
	
	{ WRITE_URB, 	0x00, 0x84 }, /* #107    seq apply     */
			
	{ WRITE_URB, 	0x10, 0x46 }, /* #108 AD_COL_BEGINL active line column = 70 */
	{ WRITE_URB, 	0x11, 0x00 }, /* #109 AD_COL_BEGINH mirror disable, gain=1 */
			
	{ WRITE_URB, 	0x00, 0x84 }, /* #110    seq apply     */
			
	{ WRITE_URB, 	0x3b, 0x0b }, /* #111       ???        */
		
	{ READ_URB, 	0x38, 0x01 }, /* #112 ---seq-----------*/
	{ WRITE_URB, 	0x38, 0x00 }, /* #113 ----- rw3--------*/
			
	{ READ_URB, 	0x38, 0x01 }, /* #114 ----seq----------*/
	{ WRITE_URB, 	0x38, 0x00 }, /* #115 ------- rw2------*/
			
    { WRITE_URB, 	0x3b, 0x0a }, /* #116-------------------*/	
	{ WRITE_URB, 	0x0e, 0x49 }, /* #117         |         */
	{ WRITE_URB, 	0x0f, 0x03 }, /* #118         |         */
	{ WRITE_URB, 	0x1c, 0x20 }, /* #119         |         */
    { WRITE_URB, 	0x1d, 0x03 }, /* #119         |         */	
	{ WRITE_URB, 	0x10, 0x46 }, /* #120         |         */
	{ WRITE_URB, 	0x11, 0x00 }, /* #121     seq write 2   */
    { WRITE_URB, 	0x14, 0x0b }, /* #122         |         */	
    { WRITE_URB, 	0x91, 0x06 }, /* #123         |         */	
    { WRITE_URB, 	0x94, 0x02 }, /* #124         |         */	
    { WRITE_URB, 	0x01, 0x29 }, /* #125         |         */	
    { WRITE_URB, 	0x91, 0x06 }, /* #126 ------------------*/	

			
	{ WRITE_URB, 	0x0c, 0xe8 }, /* #127 -----------------*/
	{ WRITE_URB, 	0x0d, 0x03 }, /* #128        |         */
	{ WRITE_URB, 	0x28, 0x82 }, /* #129        |         */ 
	{ WRITE_URB, 	0x29, 0x03 }, /* #130        |         */ 
	{ WRITE_URB, 	0x2c, 0x08 }, /* #131        |         */
	{ WRITE_URB, 	0x2d, 0x38 }, /* #132        |         */
	{ WRITE_URB, 	0x2e, 0x03 }, /* #133    seq write 1   */
	{ WRITE_URB, 	0x2f, 0x04 }, /* #134        |         */
	{ WRITE_URB, 	0x30, 0x03 }, /* #135        |         */
	{ WRITE_URB, 	0x00, 0x84 }, /* #136        |         */
	{ WRITE_URB, 	0x31, 0x01 }, /* #137 -----------------*/
			
	{ WAIT_VALUE,	0x30, 0x03 }, /* #138 - 437             */	

	{ WRITE_URB, 	0x31, 0x00 }, /* #438                   */	
	{ WRITE_URB, 	0x3b, 0x0b }, /* #439                   */	

	{ READ_URB, 	0x3a, 0xff }, /* #440 ---seq-----------*/
	{ WRITE_URB, 	0x3a, 0xff }, /* #441 ----- rw1--------*/

	{ READ_URB, 	0x38, 0x01 }, /* #442 ----seq----------*/
	{ WRITE_URB, 	0x38, 0x00 }, /* #443 ------- rw2------*/
       
    { SET_INTERFACE,0x00, 0x00 }, /* #444                   */	
    { SET_INTERFACE,0x00, 0x07 }, /* #445                   */		

	{ READ_URB, 	0x0c, 0xf4 }, /* #446 -----------------*/
	{ READ_URB, 	0x0d, 0x01 }, /* #447        |         */
	{ READ_URB, 	0x28, 0x63 }, /* #448        |         */
	{ READ_URB, 	0x29, 0x07 }, /* #449        |         */
	{ READ_URB, 	0x2c, 0x48 }, /* #450        |         */
	{ READ_URB, 	0x2d, 0x01 }, /* #451        |         */
	{ READ_URB, 	0x2e, 0x06 }, /* #452    seq read 1    */
	{ READ_URB, 	0x2f, 0x96 }, /* #453        |         */
	{ READ_URB, 	0x30, 0x04 }, /* #454        |         */
	{ READ_URB, 	0x2a, 0x73 }, /* #455        |         */
	{ READ_URB, 	0x2b, 0x73 }, /* #456        |         */
	{ READ_URB, 	0x35, 0x09 }, /* #457        |         */
	{ READ_URB, 	0x34, 0x23 }, /* #458        |         */
	{ READ_URB, 	0x83, 0x43 }, /* #459 -----------------*/

	{ WRITE_URB, 	0x10, 0x46 }, /* #460 AD_COL_BEGINL active line column = 70 */
	{ WRITE_URB, 	0x11, 0x00 }, /* #461 AD_COL_BEGINH mirror disable, gain=1 */

	{ WRITE_URB, 	0x00, 0x84 }, /* #462     seq apply     */

	{ WRITE_URB, 	0x3b, 0x0b }, /* #463                   */	
       	
	{ READ_URB, 	0x3a, 0xff }, /* #464 ---seq-----------*/
	{ WRITE_URB, 	0x3a, 0xff }, /* #465 ----- rw1--------*/

	{ READ_URB, 	0x38, 0x01 }, /* #466 ----seq----------*/
	{ WRITE_URB, 	0x38, 0x00 }, /* #467 ------- rw2------*/

	{ WRITE_URB, 	0x3b, 0x0a }, /* #468-------------------*/	
	{ WRITE_URB, 	0x0e, 0x49 }, /* #469         |         */
	{ WRITE_URB, 	0x0f, 0x03 }, /* #470         |         */
	{ WRITE_URB, 	0x1c, 0x20 }, /* #471         |         */
	{ WRITE_URB, 	0x1d, 0x03 }, /* #471         |         */	
	{ WRITE_URB, 	0x10, 0x46 }, /* #472         |         */
	{ WRITE_URB, 	0x11, 0x00 }, /* #473     seq write 2   */
	{ WRITE_URB, 	0x14, 0x0b }, /* #474         |         */	
	{ WRITE_URB, 	0x91, 0x06 }, /* #475         |         */	
	{ WRITE_URB, 	0x94, 0x02 }, /* #476         |         */	
	{ WRITE_URB, 	0x01, 0x29 }, /* #477         |         */	
	{ WRITE_URB, 	0x91, 0x06 }, /* #478 ------------------*/	

	{ WRITE_URB, 	0x00, 0x84 }, /* #479     seq apply     */
			
	{ WRITE_URB, 	0x28, 0x80 }, /* #479     RAW MODE BY   */
	{ WRITE_URB, 	0x29, 0x81 }, /* #xxx  Takafumi Mizuno  (was 0x82, 0x03) & Tomas Groth */

	{ WRITE_URB, 	0x3b, 0x0b }, /* #480                   */

	{ READ_URB, 	0x3a, 0xff }, /* #481 ---seq-----------*/
	{ WRITE_URB, 	0x3a, 0xff }, /* #482 ----- rw1--------*/

	{ READ_URB, 	0x38, 0x01 }, /* #483 ----seq----------*/
	{ WRITE_URB, 	0x38, 0x00 }, /* #484 ------- rw2------*/

    { SET_INTERFACE,0x00, 0x00 }, /* #485                   */	
    { SET_INTERFACE,0x00, 0x07 }, /* #486                   */	

	{ READ_URB, 	0x0c, 0xf4 }, /* #487 -----------------*/
	{ READ_URB, 	0x0d, 0x01 }, /* #488        |         */
	{ READ_URB, 	0x28, 0x63 }, /* #489        |         */
	{ READ_URB, 	0x29, 0x07 }, /* #490        |         */
	{ READ_URB, 	0x2c, 0x48 }, /* #491        |         */
	{ READ_URB, 	0x2d, 0x01 }, /* #492        |         */
	{ READ_URB, 	0x2e, 0x06 }, /* #493    seq read 1    */
	{ READ_URB, 	0x2f, 0x96 }, /* #494        |         */
	{ READ_URB, 	0x30, 0x04 }, /* #495        |         */
	{ READ_URB, 	0x2a, 0x73 }, /* #496        |         */
	{ READ_URB, 	0x2b, 0x73 }, /* #497        |         */
	{ READ_URB, 	0x35, 0x09 }, /* #498        |         */
	{ READ_URB, 	0x34, 0x23 }, /* #499        |         */
	{ READ_URB, 	0x83, 0x43 }, /* #500 -----------------*/	

	{ WRITE_URB, 	0x10, 0x46 }, /* #501                   */
	{ WRITE_URB, 	0x11, 0x00 }, /* #502                   */
			
	{ WRITE_URB, 	0x00, 0x84 }, /* #503     seq apply     */

	{ WRITE_URB, 	0x3b, 0x0b }, /* #504                   */
			
	{ READ_URB, 	0x3a, 0xff }, /* #505 ---seq-----------*/
	{ WRITE_URB, 	0x3a, 0xff }, /* #506 ----- rw1--------*/
			
	{ READ_URB, 	0x38, 0x01 }, /* #507 ---seq-----------*/
	{ WRITE_URB, 	0x38, 0x00 }, /* #508 ----- rw3--------*/
			
    { WRITE_URB, 	0x3b, 0x0a }, /* #509-------------------*/	
	{ WRITE_URB, 	0x0e, 0x49 }, /* #469         |         */
	{ WRITE_URB, 	0x0f, 0x03 }, /* #470         |         */
	{ WRITE_URB, 	0x1c, 0x20 }, /* #471         |         */
	{ WRITE_URB, 	0x1d, 0x03 }, /* #471         |         */	
	{ WRITE_URB, 	0x10, 0x46 }, /* #513         |         */
	{ WRITE_URB, 	0x11, 0x00 }, /* #514     seq write 2   */
    { WRITE_URB, 	0x14, 0x0b }, /* #515         |         */	
	{ WRITE_URB, 	0x91, 0x06 }, /* #516         |         */
    { WRITE_URB, 	0x94, 0x02 }, /* #517         |         */	
    { WRITE_URB, 	0x01, 0x29 }, /* #518         |         */	
    { WRITE_URB, 	0x91, 0x06 }, /* #519 ------------------*/	

	{ WRITE_URB, 	0x00, 0x84 }, /* #520     seq apply     */

	{ RESET_PIPE, 	0x00, 0x81 }, /* #521     reset pipe     */
};

/*
 * ---------------------------------------------------------
 *                  Camera control stuff
 * ---------------------------------------------------------
 */
struct icm532_camera {
	bool is_opened;
	bool is_removed;
	bool is_initialized;
	bool is_capturing;
	bool have_to_stop;
	u8 *raw_line;		// raw data captured from the camera
	u8 *framebuf;		// processed data in RGB24 format
	int decoded_lines;
	u32 framebuf_size;	// # of valid bytes in framebuf
	
	int raw_in_progress;
	int raw_last_not_empty;
	int raw_size[2];
	int raw_ready[2];
	int raw_requested[2];
	
	u8 sequence;
	
	int urb_sequence;
	struct urb *start_urb[2];	// pointer to allocated urb structure for isoc transfers
	struct urb *iso_urb[10];
	int number;
	int urbs_onair;
	u8 *urb_data;
	
	struct video_device vdev;	// v4l video device
	struct usb_device *udev;	// usb device
	wait_queue_head_t wait_queue;	// wait queue

};



static int write_byte_register (struct usb_device *dev, u8 icm_register, u8 value);

static int read_byte_register (struct usb_device *dev, u8 icm_register, u8 *value);

int urb_message(struct icm532_camera *cam, icm532_urb_t message);

static int initialize_camera(struct icm532_camera *cam);

static int start_capture(struct icm532_camera *cam);

static int stop_capture(struct icm532_camera *cam);

static int urb_iso_sequence(struct icm532_camera *cam);

static void urb_iso_complete(struct urb *urb);

void new_bayer2rgb(unsigned char *dst, unsigned char *src);

void stop_grab(struct icm532_camera *cam);

// ---------------------------------------------------------
//                  Video Device stuff
// ---------------------------------------------------------

static int
icm532_open(struct video_device *dev, int flags);		

static void
icm532_close(struct video_device *dev);		

static long
icm532_read(struct video_device *dev, char *buf,
		 unsigned long count, int noblock);

static long
icm532_write(struct video_device *v, const char *buf, unsigned long count,
		  int nonblock);

static int
icm532_ioctl(struct video_device *dev, unsigned int ioctlnr, void *arg);


static int
icm532_mmap(struct video_device *dev, const char *adr, unsigned long size);


int
icm532_video_init(struct video_device *vdev);


static struct video_device icm532_video_device = {
	owner:		THIS_MODULE,
	name:		"ICM532 based USB Camera",
	type:		VID_TYPE_CAPTURE,
	hardware:	VID_HARDWARE_VICAM,
	open:		icm532_open,
	close:		icm532_close,
	read:		icm532_read,
	write:		icm532_write,
	ioctl:		icm532_ioctl,
	mmap:		icm532_mmap,
	initialize:	icm532_video_init,
	minor:		-1,
};

static void
read_frame(struct icm532_camera *cam, int framenum);

// -----------------------------------------------------------
//           USB Driver stuff
// -----------------------------------------------------------

/* Define these values to match your device */
#define USB_ICM532_VENDOR_ID	0x0923
#define USB_ICM532_PRODUCT_ID	0x010f


static struct usb_device_id icm532_table[] = {
	{USB_DEVICE(USB_ICM532_VENDOR_ID, USB_ICM532_PRODUCT_ID)},
	{}
};

MODULE_DEVICE_TABLE(usb, icm532_table);

static void *
icm532_probe(struct usb_device *dev, unsigned int ifnum,
		  const struct usb_device_id *id);

static void
icm532_purge(struct icm532_camera *cam);

static void
icm532_disconnect(struct usb_device *dev, void *ptr);

static struct usb_driver icm532_driver = {
	name:		"icm532 based cam",
	probe:		icm532_probe,
	disconnect:	icm532_disconnect,
	id_table:	icm532_table,
};






// -----------------------------------------------------------
//           Kernel module stuff
// -----------------------------------------------------------

static int icm532_init(void);

static void icm532_exit(void);

module_init(icm532_init);

module_exit(icm532_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);

MODULE_DESCRIPTION(DRIVER_DESC);

MODULE_LICENSE("GPL");
